#include <stdio.h>
#include <stdlib.h>

unsigned char byteswap;
unsigned int pc; // for printheader

void
byteswapped(FILE *f)
{
  unsigned int data;
  fseek(f, 0, SEEK_SET);
  fread(&data, 1, 4, f);
  
  if (data == 0x80371240) // No byteswapping
    byteswap = 0;
  else if (data == 0x37804012) // 1 byte swapping
    byteswap = 1;
  else if (data == 0x12408037) // 2 byte swapping
    byteswap = 2;
  else if (data == 0x40123780) // 2, then 1 byte swapping (unlikely)
    byteswap = 3;
  else {
    fprintf(stderr, "Unknown byteswapping encountered, or invalid file.\n");
    fprintf(stderr, "recieved 0x%08x as validation word.\n", data);
    fprintf(stderr, "if you see a 13 in there, it may be that the file is compressed\n");
    exit(1);
  }
}

void
byteswapread(unsigned int *data, unsigned int size, FILE *f)
{
  unsigned char c[size];
  unsigned int i;
  
  fread(c, size, 1, f);

  for (i = 0 ; i < size/4 ; i++) {
    if (byteswap == 1) {
      data[i] =
	(c[(i*4)+2] << 24) | (c[(i*4)+3] << 16) |
	(c[(i*4)] << 8) | c[(i*4)+1];
    } else if (byteswap == 2) {
      data[i] =
	(c[(i*4)+1] << 24) | (c[(i*4)] << 16) |
	(c[(i*4)+3] << 8) | c[(i*4)+2];
    } else if (byteswap == 3) {
      data[i] =
	(c[(i*4)] << 24) | (c[(i*4)+1] << 16) |
	(c[(i*4)+2] << 8) | c[(i*4)+3];
    }
  }
}

void
printheader(FILE *f)
{
  unsigned int ints[0x10];
  unsigned char chars[0x40];
  unsigned int i;
  
  fseek(f, 0, SEEK_SET);
  byteswapread(ints, 0x40, f);

  for (i = 0 ; i < 0x10 ; i++) {
    chars[(i*4)]  =  (ints[i] & 0xff000000) >> 24;
    chars[(i*4)+1] = (ints[i] & 0x00ff0000) >> 16;
    chars[(i*4)+2] = (ints[i] & 0x0000ff00) >> 8;
    chars[(i*4)+3] = (ints[i] & 0x000000ff);
  }

  printf(".romname \"%s\"\n", chars + 0x20);

  printf(".country %s\n", (chars[0x3e] == '\0') ? "None" :
	 (chars[0x3e] == 'D') ? "Germany" :
	 (chars[0x3e] == 'E') ? "USA" :
	 (chars[0x3e] == 'J') ? "Japan" :
	 (chars[0x3e] == 'P') ? "Europe" :
	 (chars[0x3e] == 'U') ? "Austrailia" : "Unknown");

  printf(".manufacturer %s\n", (chars[0x3b] == '\0') ? "None" :
	 (chars[0x3b] == 'N') ? "Nintendo" : "Unknown");

  pc = ints[2];
  printf(".pc 0x%08x\n\n", pc);
}
