#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <obstack.h>
#include "nop.h"

unsigned int pc = 0x80200000;
unsigned int numfunction;

void
readargs(char **filename, long *offset, int argc, char **argv)
{
  int i;
  unsigned int addr;
  mipsOp op;
  
  *filename[0] = '\0';
  for (i = 1 ; i < argc ; i++) {
    if (argv[i][0] == '-') {
      switch(argv[i][1]) {
      case 'f':
	if (argv[i][2] == '\0') {
	  i++;
	  strcpy(*filename, argv[i]);
	} else {
	  strcpy(*filename, argv[i]+2);
	}
	break;
      case 'o':
	if (argv[i][2] == '\0') {
	  char *err;
	  i++;
	  *offset = strtol(argv[i], &err, 0);
	  if ( (*argv[i] == '\0') || (*err != '\0') )
	    fprintf(stderr, "There was a problem with the argument: %s\n",
		    argv[i]);
	} else {
	  char *err;
	  *offset = strtol(argv[i]+2, &err, 0);
	  if ( (*argv[i]+2 == '\0') || (*err != '\0') )
	    fprintf(stderr, "There was a problem with the argument: %s\n",
		    argv[i]+2);
	}
	break;
      case 'd':
	if (argv[i][2] == '\0') {
	  char *err;
	  i++;
	  op.data = strtoul(argv[i], &err, 0);
	  if ( (*argv[i] == '\0') || (*err != '\0') )
	    fprintf(stderr, "There was a problem with the argument: %s\n",
		    argv[i]);
	} else {
	  char *err;
	  op.data = strtoul(argv[i]+2, &err, 0);
	  if ( (*argv[i]+2 == '\0') || (*err != '\0') )
	    fprintf(stderr, "There was a problem with the argument: %s\n",
		    argv[i]+2);
	}
	fprintf(stderr, "OP:     %02d\n", op.f.op);
	fprintf(stderr, "RS:     %02d\n", op.f.rs);
	fprintf(stderr, "RT:     %02d\n", op.f.rt);
	fprintf(stderr, "RD:     %02d\n", op.f.rd);
	fprintf(stderr, "SA:     %02d\n", op.f.sa);
	fprintf(stderr, "Func:   %02d\n", op.f.func);
	fprintf(stderr, "Imm:    %02u\n", op.i.imm);
	fprintf(stderr, "Off:    %02d\n", ((signed char)op.i.imm << 2) / 4);
	fprintf(stderr, "Opcode: %s\n", getopcode(op));
	exit(0);
	break;
      case 'u':
	if (argv[i][2] == '\0') {
	  char *err;
	  i++;
	  addr = strtoul(argv[i], &err, 0);
	  if ( (*argv[i] == '\0') || (*err != '\0') )
	    fprintf(stderr, "There was a problem with the argument: %s\n",
		    argv[i]);
	} else {
	  char *err;
	  addr = strtoul(argv[i]+2, &err, 0);
	  if ( (*argv[i]+2 == '\0') || (*err != '\0') )
	    fprintf(stderr, "There was a problem with the argument: %s\n",
		    argv[i]+2);
	}
	addfunction(addr);
	break;
      case 'h':
	fprintf(stderr, "Nop (v1.1) Help:\n");
	fprintf(stderr, "\t-f<filename>\tuse specified file\n");
	fprintf(stderr, "\t-o<offset>\tstart loading from specified offset\n");
	fprintf(stderr, "\t-d<val>\t\tdisasm val\n");
	fprintf(stderr, "\t-u<val>\t\tthere's a function at memory address val\n");
	fprintf(stderr, "\t\t\tcan be used any number of times, first use is func1\n");
	fprintf(stderr, "\nProblems? Mail Tom Whittock <shallow@dial.pipex.com>\n");
	exit(0);
	break;
      default:
	fprintf(stderr, "Unknown argument: %s\n", argv[i]);
	break;
      }
    } else {
      fprintf(stderr, "Unknown argument: %s\n", argv[i]);
    }
  }

  *filename = realloc(*filename, strlen(*filename)+1);
}

void
printop(Instr inst)
{
  optype OPCODETYPE = getoptype(inst.op);
  const char *OPCODE = getopcode(inst.op);

  if (inst.what == LABEL) printf("$L%d:\t", inst.info);
  else printf("\t");
  
  if (inst.op.data == 0) printf("nop\n");
  else {
    switch (OPCODETYPE) {
    case OP:
      printf("%s\n", OPCODE);
      break;
    case OPTGT:
      printf("%s\tfunc%d\n", OPCODE, inst.info);
      break;
    case OPC:
      printf("%s\t0x%x\n", OPCODE, inst.op.c.code);
      break;
    case OPS:
      printf("%s\t%s\n", OPCODE, GPR[inst.op.f.rs]);
      break;
    case OPSD:
      printf("%s\t%s\t%s\n", OPCODE, GPR[inst.op.f.rs], GPR[inst.op.f.rd]);
      break;
    case OPST:
      printf("%s\t%s\t%s\n", OPCODE, GPR[inst.op.f.rs], GPR[inst.op.f.rt]);
      break;
    case OPSI:
      printf("%s\t%s\t0x%x\n", OPCODE, GPR[inst.op.f.rs], inst.op.i.imm);
      break;
    case OPSO:
      printf("%s\t%s\t$L%d\n", OPCODE, GPR[inst.op.f.rs], inst.info);
      break;
    case OPSTO:
      printf("%s\t%s\t%s\t$L%d\n", OPCODE, GPR[inst.op.f.rs], GPR[inst.op.f.rt], inst.info);
      break;
    case OPTI:
      printf("%s\t%s\t0x%x\n", OPCODE, GPR[inst.op.f.rt], inst.op.i.imm);
      break;
    case OPTSI:
      printf("%s\t%s\t%s\t0x%x\n", OPCODE, GPR[inst.op.f.rt], GPR[inst.op.f.rs], inst.op.i.imm);
      break;
    case OPTOB:
      printf("%s\t%s\t0x%x(%s)\n", OPCODE, GPR[inst.op.f.rt], inst.op.i.imm, GPR[inst.op.f.rs]);
      break;
    case OPTD:
      printf("%s\t%s\t%s\n", OPCODE, GPR[inst.op.f.rt], GPR[inst.op.f.rd]);
      break;
    case OPD:
      printf("%s\t%s\n", OPCODE, GPR[inst.op.f.rd]);
      break;
    case OPDST:
      printf("%s\t%s\t%s\t%s\n", OPCODE, GPR[inst.op.f.rd], GPR[inst.op.f.rs], GPR[inst.op.f.rt]);
      break;
    case OPDTA:
      printf("%s\t%s\t%s\t%d\n", OPCODE, GPR[inst.op.f.rd], GPR[inst.op.f.rt], inst.op.f.sa);
      break;
    case OPDTS:
      printf("%s\t%s\t%s\t%s\n", OPCODE, GPR[inst.op.f.rd], GPR[inst.op.f.rt], GPR[inst.op.f.rs]);
      break;
    case OPTSD:
      printf("%s\t%s\t%s\n", OPCODE, GPR[inst.op.f.rt], SYSR[inst.op.f.rd]);
      break;
    case OPTSS:
      printf("%s\t%s\t%s\n", OPCODE, GPR[inst.op.f.rt], SYSR[inst.op.f.rs]);
      break;
    case OPTFD:
      printf("%s\t%s\t$f%02d\n", OPCODE, GPR[inst.op.f.rt], inst.op.f.rd);
      break;
    case OPTFS:
      printf("%s\t%s\t$f%02d\n", OPCODE, GPR[inst.op.f.rt], inst.op.f.rs);
      break;
    case OPFTOB:
      printf("%s\t$f%02d\t0x%x(%s)\n", OPCODE, inst.op.f.rt, inst.op.i.imm, GPR[inst.op.f.rs]);
      break;
    case OPRTOB:
      printf("%s\t$v%02d\t0x%x(%s)\n", OPCODE, inst.op.f.rt, inst.op.i.imm, GPR[inst.op.f.rs]);
      break;
    case OPFSFT:
      printf("%s\t$f%02d\t$f%02d\n", OPCODE, inst.op.f.rs, inst.op.f.rt);
      break;
    case OPFDFS:
      printf("%s\t$f%02d\t$f%02d\n", OPCODE, inst.op.f.rd, inst.op.f.rs);
      break;
    case OPFDFSFT:
      printf("%s%s\t$f%02d\t$f%02d\t$f%02d\n", OPCODE, cop1Ops[inst.op.f.rs], inst.op.f.rd, inst.op.f.rs, inst.op.f.rt);
      break;
    case OPTRD:
      printf("%s\t%s\t$v%02d\n", OPCODE, GPR[inst.op.f.rt], inst.op.f.rd);
      break;
    case OPTRS:
      printf("%s\t%s\t$v%02d\n", OPCODE, GPR[inst.op.f.rt], inst.op.f.rs);
      break;
    case COPBC:
      printf("%s%c%s\t0x%x\n", OPCODE,
	     (inst.op.bc.tf == 1) ? 't' : 'f',
	     (inst.op.bc.nd == 1) ? "l" : "", inst.op.bc.imm);
      break;
    case TLB:
      switch (inst.op.f.func) {
      case 1:
	printf("tlbr\n");
	break;
      case 2:
	printf("tlbwi\n");
	break;
      case 6:
	printf("tlbwr\n");
	break;
      case 8:
	printf("tlbp\n");
	break;
      case 24:
	printf("eret\n");
	break;
      default:
	printf("Unknown Cop0 TLB Instruction\n");
	break;
      }
      break;
    case VECTOR:
      printf("%s\t$v%02d\t$v%02d\t$v%02d[%02d]\t;VECTOR\n", OPCODE, inst.op.v.dest, inst.op.v.s1, inst.op.v.s2, inst.op.v.el);
      break;
    case HELP:
      printf("%s\t[Unknown Format - Help]\n", OPCODE);
      break;
    case SPECIAL:
    case REGIMM:
    case FPU:
    case COP0:
    case COP1:
    case COP2:
    case NONE:
      printf("Invalid or Unknown opcode: 0x%08x\n", inst.op.data);
      break;
    }
  }
}

void
readfunc(FILE *f, unsigned int funcnum)
{
  typedef struct {
    unsigned int num;
    unsigned int to;
  } labelinf;

  Instr insr;
  Instr *insrs;
  labelinf label;
  labelinf *labs;
  
  struct obstack func;
  struct obstack labels;

  unsigned int i;
  unsigned int loclabelnum = 1;
  unsigned int labstack = 0;
  unsigned int funcsize;
  unsigned int minfuncsize = 0;
  char onemore = 0;
  int OPCODETYPE;

  obstack_init(&func);
  obstack_init(&labels);

  label.to = 0;
  label.num = 0;
  obstack_grow(&labels, &label, sizeof(labelinf));
  
  for (funcsize = 0 ;; funcsize++) {
    if (feof(f)) {
      fprintf(stderr, "Error disassembling function: EOF reached.\n");
      fprintf(stderr, "This means that function %d is probably useless\n",
	      funcnum);
      fprintf(stderr, "This also indicates a bug in Nop.\n");
      break;
    }
    byteswapread(&insr.op.data, 4, f);

    OPCODETYPE = getoptype(insr.op);

    if ( (OPCODETYPE == OPSO) || (OPCODETYPE == OPSTO) ) {
      char done = 0;
      label.to = funcsize + 1 + ((signed char)insr.op.i.imm << 2) / 4;

      labs = (labelinf *)obstack_base(&labels);
      for (i = 1 ; i < loclabelnum ; i++) {
	if (labs[i].to == label.to) {
	  insr.what = BRANCH;
	  insr.info = labs[i].num;
	  done = 1;
	  break;
	}
      }

      if (!done) {
	if (label.to > minfuncsize)
	  minfuncsize = label.to;
	
	label.num = loclabelnum;
      
	obstack_grow(&labels, (void *)&label, sizeof(labelinf));
      
	insr.what = BRANCH;
	insr.info = loclabelnum;

	loclabelnum++;
      }
    } else if (OPCODETYPE == OPTGT) {
      unsigned int t =
	addfunction((((unsigned int) (insr.op.t.target << 2)) | pc) - pc);
      insr.what = JUMP;
      insr.info = t;
    } else {
      insr.what = NOTHING;
    }

    labs = (labelinf *) obstack_base(&labels);
    for (i = 0 ; i < labstack ; i++) {
      if (labs[i].to == funcsize) {
	insr.what = LABEL;
	insr.info = labs[i].num;
	break;
      }
    }

    obstack_grow(&func, &insr, sizeof(Instr));

    if (funcsize >= minfuncsize) { // so we can jump past these
      if (onemore == 1) break;
      else if (insr.op.f.op == 2)
	onemore = 1; // 'j'
      else if ( (insr.op.f.op == 0)&&(insr.op.f.func == 8) )
	onemore = 1; // 'jr'
      else if ( (insr.op.f.op == 0)&&(insr.op.f.func == 13) )
	break; // 'break'
    }
  }

  insrs = obstack_finish(&func);
  labs = obstack_finish(&labels);
  
  printf("func%d:\n", funcnum);
  for (i = 0 ; i <= funcsize ; i++) {
    unsigned int j;

    for (j = 1 ; j < loclabelnum ; j++) {
      if (labs[j].to == i) {
	insrs[i].what = LABEL;
	insrs[i].info = labs[j].num;
	break;
      }
    }

    printop(insrs[i]);
  }
  printf("\n");

  obstack_free(&func, insrs);
  obstack_free(&labels, labs);
}

int
main(int argc, char **argv)
{
  char *filename = malloc(256);
  long offset = 0x1000;
  FILE *f;
  unsigned int i;
  
  function_init();

  readargs(&filename, &offset, argc, argv);

  if (*filename == '\0') {
    fprintf(stderr, "No filename given: use -f\n");
    free(filename);
    function_free();
    return 1;
  }
  
  f = fopen(filename, "rb");
  if (f == NULL) {
    fprintf(stderr, "Couldn't open file %s\n", filename);
    perror("nop");
    free(filename);
    function_free();
    return 1;
  }

  byteswapped(f);

  printheader(f); // updates pc

  for(i = 0 ; i < numfunction ; i++) {
    fseek(f, getfuncaddr(i) + offset, SEEK_SET);
    readfunc(f, i);
  }
  
  fclose(f);
  fflush(stdout);
  free(filename);
  function_free();
  return 0;
}
