#ifndef __NOP_H__
#define __NOP_H__

typedef enum { /* All the opcodes in the world - sort of */
  NONE,
  HELP,     /* Eh-oh! */
  SPECIAL,
  REGIMM,
  VECTOR,
  FPU,
  COP0,
  COP1,
  COP2,     /* Below: Try to work out my system! */
  OP,       /* Format: Opcode */
  OPTGT,    /* Format: Opcode Target */
  OPC,      /* Format: Opcode Code (System Call, Break, ?Sync?) */
  OPS,      /* Format: Opcode $RS */
  OPSD,     /* Format: Opcode $RS $RD */
  OPST,     /* Format: Opcode $RS $RT */
  OPSI,     /* Format: Opcode $RS Immediate */
  OPSO,     /* Format: Opcode $RS Offset */
  OPSTO,    /* Format: Opcode $RS $RT Offset */
  OPTI,     /* Format: Opcode $RT Immediate */
  OPTSI,    /* Format: Opcode $RT $RS Immediate */
  OPTOB,    /* Format: Opcode $RT Offset(base) */
  OPTD,     /* Format: Opcode $RT $RD */
  OPD,      /* Format: Opcode $RD */
  OPDST,    /* Format: Opcode $RD $RS $RT */
  OPDTA,    /* Format: Opcode $RD $RT ShiftAmount */
  OPDTS,    /* Format: Opcode $RD $RT $RS */
  OPTSD,    /* Format: Opcode $RT $SD (SD = System Reg) */
  OPTSS,    /* Format: Opcode $RT $SS (SS = System Reg) */
  OPTFD,    /* Format: Opcode $RT $FD (FD = FPU Reg) */
  OPTFS,    /* Format: Opcode $RT $FS (FS = FPU Reg) */
  OPFTOB,   /* Format: Opcode $FT Offset(base) (FT = FPU Reg) */
  OPRTOB,   /* Format: Opcode $RT Offset(base) (RT = RCP Reg) */
  OPFSFT,   /* Format: Opcode $FS $FT */
  OPFDFS,   /* Format: Opcode $FD $FS */
  OPFDFSFT, /* Format: Opcode $FD $FS $FT */
  OPTRD,    /* Format: Opcode $RT $VD (VD = Vector Reg) */
  OPTRS,    /* Format: Opcode $RT $VS (VS = Vector Reg) */
  COPBC,    /* Coprocessor Branch - see mipsOp.bc */
  TLB       /* Translation Lookup Buffer? Lookaside? Something like that */
} optype;

typedef enum {
  NOTHING,
  BRANCH,
  JUMP,
  LABEL
} whatisit;

typedef union
{
  struct
  {
    unsigned func:6;
    unsigned sa:5;
    unsigned rd:5;
    unsigned rt:5;
    unsigned rs:5;
    unsigned op:6;
  } f;
  struct
  {
    unsigned func:6;
    unsigned code:20;
    unsigned op:6;
  } c;
  struct
  {
    unsigned imm:16;
    unsigned rt:5;
    unsigned rs:5;
    unsigned op:6;
  } i;
  struct
  {
    unsigned target:26;
    unsigned op:6;
  } t;
  struct 
  {
    unsigned imm:16;
    unsigned tf:1;
    unsigned nd:1;
    unsigned pad:3;
    unsigned rs:5;
    unsigned op:6;
  } bc;
  struct 
  {
    unsigned pad:6;
    unsigned dest:5;
    unsigned s1:5;
    unsigned s2:5;
    unsigned el:4;
    unsigned one:1;
    unsigned op:6;
  } v;
  unsigned int data;
  struct
  {
    unsigned b1:8;
    unsigned b2:8;
    unsigned b3:8;
    unsigned b4:8;
  } byte;
  struct
  {
    unsigned w1:16;
    unsigned w2:16;
  } word;
} mipsOp;

typedef struct {
  mipsOp op;
  unsigned int info;
  whatisit what;
} Instr;

typedef struct {
  unsigned int addr;
  unsigned int size;
} Function;

static const char *GPR[32] = { /* Main Processor Regs */
  "$zero", "$at", "$v0", "$v1", "$a0", "$a1", "$a2", "$a3",
  "$t0",   "$t1", "$t2", "$t3", "$t4", "$t5", "$t6", "$t7",
  "$s0",   "$s1", "$s2", "$s3", "$s4", "$s5", "$s6", "$s7",
  "$t8",   "$t9", "$k0", "$k1", "$gp", "$sp", "$s8", "$ra"
};

static const char *SYSR[32] = { /* Cop0 Regs - System Regs*/
  "$Index", "$Random", "$EntryLo0", "$EntryLo1", "$Context", "$PageMask", "$Wired", "$C0-07",
  "$BadVAddr", "$Count", "$EntryHi", "$Compare", "$Status", "$Cause", "$EPC", "$PRevID",
  "$Config", "$LLAddr", "$WatchLo", "$WatchHi", "$XContext", "$C0-21", "$C0-22", "$C0-23",
  "$C0-24", "$C0-25", "$PErr", "$CacheErr", "$TagLo", "$TagHi", "$ErrorEPC", "$C0-31"
};

static const char *mainOps[64] = {
  "special", "regimm", "j", "jal", "beq", "bne", "blez", "bgtz",
  "addi", "addiu", "slti", "sltiu", "andi", "ori", "xori", "lui",
  "cop0", "cop1", "cop2", "", "beql", "bnel", "blezl", "bgtzl",
  "daddi", "daddiu", "ldl", "ldr", "", "", "", "",
  "lb", "lh", "lwl", "lw", "lbu", "lhu", "lwr", "lwu",
  "sb", "sh", "swl", "sw", "sdl", "sdr", "swr", "cache",
  "ll", "lwc1", "lwc2", "", "lld", "ldc1", "ldc2", "ld",
  "sc", "swc1", "swc2", "", "scd", "sdc1", "sdc2", "sd"
};

static const optype mainOpsType[64] = {
  SPECIAL, REGIMM, OPTGT,  OPTGT, OPSTO, OPSTO,  OPSO,   OPSO,
  OPTSI,   OPTSI,  OPTSI,  OPTSI, OPTSI, OPTSI,  OPTI,   OPTI,
  COP0,    COP1,   COP2,   NONE,  OPSTO, OPSTO,  OPSO,   OPSO,
  OPTSI,   OPTSI,  OPTOB,  OPTOB, NONE,  NONE,   NONE,   NONE,
  OPTOB,   OPTOB,  OPTOB,  OPTOB, OPTOB, OPTOB,  OPTOB,  OPTOB,
  OPTOB,   OPTOB,  OPTOB,  OPTOB, OPTOB, OPTOB,  OPTOB,  OPTOB,
  OPTOB,   OPFTOB, OPRTOB, NONE,  OPTOB, OPFTOB, OPRTOB, OPTOB,
  OPTOB,   OPFTOB, OPRTOB, NONE,  OPTOB, OPFTOB, OPRTOB, OPTOB
};

static const char *specialOps[64] = {
  "sll", "", "srl", "sra", "sllv", "", "srlv", "srav",
  "jr", "jalr", "", "", "syscall", "break", "", "sync",
  "mfhi", "mthi", "mflo", "mtlo", "dsllv", "", "dsrlv", "dsrav",
  "mult", "multu", "div", "divu", "dmult", "dmultu", "ddiv", "ddivu",
  "add", "addu", "sub", "subu", "and", "or", "xor", "nor",
  "", "", "slt", "sltu", "dadd", "daddu", "dsub", "dsubu",
  "tge", "tgeu", "tlt", "tltu", "teq", "", "tne", "",
  "dsll", "", "dsrl", "dsra", "dsll32", "", "dsrl32", "dsra32"
};

static const optype specialOpsType[64] = {
  OPDTA, NONE,  OPDTA, OPDTA, OPDTS, NONE,  OPDTS, OPDTS,
  OPS,   OPSD,  NONE,  NONE,  OPC,   OPC,   NONE,  OP,
  OPD,   OPD,   OPD,   OPD,   OPDTS, NONE,  OPDTS, OPDTS,
  OPST,  OPST,  OPST,  OPST,  OPST,  OPST,  OPST,  OPST,
  OPDST, OPDST, OPDST, OPDST, OPDST, OPDST, OPDST, OPDST,
  NONE,  NONE,  OPDST, OPDST, OPDST, OPDST, OPDST, OPDST,
  OPST,  OPST,  OPST,  OPST,  OPST,  NONE,  OPST,  NONE,
  OPDTA, NONE,  OPDTA, OPDTA, OPDTA, NONE,  OPDTA, OPDTA
};

static const char *regimmOps[32] = {
  "bltz", "bgez", "bltzl", "bgezl", "", "", "", "",
  "tgei", "tgeiu", "tlti", "tltiu", "teqi", "", "tnei", "",
  "bltzal", "bgezal", "bltzall", "bgezall", "", "", "", "",
  "", "", "", "", "", "", "", ""
};

static const optype regimmOpsType[32] = {
  OPSO, OPSO, OPSO, OPSO, NONE, NONE, NONE, NONE,
  OPSI, OPSI, OPSI, OPSI, OPSI, NONE, OPSI, NONE,
  OPSO, OPSO, OPSO, OPSO, NONE, NONE, NONE, NONE,
  NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE
};

static const char *cop0Ops[32] = {
  "mfc0", "dmfc0", "cfc0", "", "mtc0", "dmtc0", "ctc0", "",
  "bc0", "", "", "", "", "", "", "",
  "tlb", "", "", "", "", "", "", "",
  "", "", "", "", "", "", "", ""
};

static const optype cop0OpsType[32] = {
  OPTSD, OPTSD, OPTSS, NONE, OPTSD, OPTSD, OPTSS, NONE,
  COPBC, NONE,  NONE,  NONE, NONE,  NONE,  NONE,  NONE,
  TLB,   NONE,  NONE,  NONE, NONE,  NONE,  NONE,  NONE,
  NONE,  NONE,  NONE,  NONE, NONE,  NONE,  NONE,  NONE
};

static const char *cop1Ops[32] = {
  "mfc1", "dmfc1", "cfc1", "", "mtc1", "dmtc1", "ctc1", "",
  "bc1", "", "", "", "", "", "", "",
  ".s", ".d", "", "", ".w", ".l", "", "",
  "", "", "", "", "", "", "", ""
};

static const optype cop1OpsType[32] = {
  OPTFD, OPTFD, OPTFS, NONE, OPTFD, OPTFD, OPTFS, NONE,
  COPBC, NONE,  NONE,  NONE, NONE,  NONE,  NONE,  NONE,
  FPU,   FPU,   NONE,  NONE, FPU,   FPU,   NONE,  NONE,
  NONE,  NONE,  NONE,  NONE, NONE,  NONE,  NONE,  NONE
};

static const char *cop1FPUOps[64] = {
  "add", "sub", "mul", "div", "sqrt", "abs", "mov", "neg",
  "round.l", "trunc.l", "ceil.l", "floor.l", "round.w", "trunc.w", "ceil.w", "floor.w",
  "", "", "", "", "", "", "", "",
  "", "", "", "", "", "", "", "",
  "cvt.s", "cvt.d", "", "", "cvt.w", "cvt.l", "", "",
  "", "", "", "", "", "", "", "",
  "c.f", "c.un", "c.eq", "c.ueq", "c.olt", "c.ult", "c.ole", "c.ule",
  "c.sf", "c.ngle", "c.seq", "c.ngl", "c.lt", "c.nge", "c.le", "c.ngt"
};

static const optype cop1FPUOpsType[64] = {
  OPFDFSFT, OPFDFSFT, OPFDFSFT, OPFDFSFT, OPFDFS, OPFDFS, OPFDFS, OPFDFS,
  OPFDFS, OPFDFS, OPFDFS, OPFDFS, OPFDFS, OPFDFS, OPFDFS, OPFDFS,
  NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,
  NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,
  OPFDFS, OPFDFS, NONE, NONE, OPFDFS, OPFDFS, NONE, NONE,
  NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,
  OPFSFT, OPFSFT, OPFSFT, OPFSFT, OPFSFT, OPFSFT, OPFSFT, OPFSFT,
  OPFSFT, OPFSFT, OPFSFT, OPFSFT, OPFSFT, OPFSFT, OPFSFT, OPFSFT
};

static const char *cop2Ops[32] = {
  "mfc2", "dmfc2", "cfc2", "", "mtc2", "dmtc2", "ctc2", "",
  "", "", "", "", "", "", "", "",
  "", "", "", "", "", "", "", "",
  "", "", "", "", "", "", "", ""
};

static const optype cop2OpsType[32] = {
  OPTRD,  OPTRD,  OPTRS,  NONE,   OPTRD,  OPTRD,  OPTRS,  NONE,
  COPBC,  NONE,   NONE,   NONE,   NONE,   NONE,   NONE,   NONE,
  VECTOR, VECTOR, VECTOR, VECTOR, VECTOR, VECTOR, VECTOR, VECTOR,
  VECTOR, VECTOR, VECTOR, VECTOR, VECTOR, VECTOR, VECTOR, VECTOR
};

static const char *cop2LoadOps[32] = {
  "lbv", "lsv", "llv", "ldv", "lqv", "lrv", "lpv", "luv",
  "lhv", "lfv", "lwv", "ltv", "", "", "", "",
  "", "", "", "", "", "", "", "",
  "", "", "", "", "", "", "", ""
};

static const char *cop2StoreOps[32] = {
  "sbv", "ssv", "slv", "sdv", "sqv", "srv", "spv", "suv",
  "shv", "sfv", "swv", "stv", "", "", "", "",
  "", "", "", "", "", "", "", "",
  "", "", "", "", "", "", "", ""
};

static const char *cop2VectorOps[64] = {
  "vmulf", "vmulu", "vrndp", "vmulq", "vmudl", "vmudm", "vmudn", "vmudh",
  "vmacf", "vmacu", "vrndn", "vmacq", "vmadl", "vmadm", "vmadn", "vmadh",
  "vadd", "vsub", "", "vabs", "addc", "vsubc", "", "",
  "", "", "", "", "", "", "", "",
  "vlt", "veq", "vne", "vge", "vcl", "vch", "vcr", "vmrg",
  "vand", "vnand", "vor", "vnor", "vxor", "vnxor", "", "",
  "vrcp", "vrcpl", "vrcph", "vmov", "vrsq", "vrsql", "vrsqh",
  "", "", "", "", "", "", "", ""
};

static const optype cop2VectorOpsType[64] = {
  VECTOR, VECTOR, VECTOR, VECTOR, VECTOR, VECTOR, VECTOR, VECTOR,
  HELP, HELP, NONE, HELP, HELP, HELP, NONE, NONE,
  NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,
  HELP, HELP, HELP, HELP, HELP, HELP, HELP, HELP,
  HELP, HELP, HELP, HELP, HELP, HELP, NONE, NONE,
  HELP, HELP, HELP, HELP, HELP, HELP, HELP, HELP,
  NONE, NONE, NONE, NONE, NONE, NONE, NONE, NONE,
};

void byteswapped(FILE *f);
void byteswapread(unsigned int *data, unsigned int size, FILE *f);
void printheader(FILE *f);

unsigned int addfunction(unsigned int addr);
unsigned int getfuncaddr(unsigned int i);
void updatefunctionaddr(unsigned int i, unsigned int addr);
void function_init(void);
void function_free(void);

#define obstack_chunk_alloc malloc
#define obstack_chunk_free free

#define getoptype(mipsOp) \
      (mainOpsType[mipsOp.f.op] == SPECIAL) ? specialOpsType[mipsOp.f.func] : \
(mainOpsType[mipsOp.f.op] == REGIMM) ? regimmOpsType[mipsOp.f.rt] : \
      (mainOpsType[mipsOp.f.op] == COP0) ? cop0OpsType[mipsOp.f.rs] : \
      (mainOpsType[mipsOp.f.op] == COP1) ? ((cop1OpsType[mipsOp.f.rs] == FPU) ? cop1FPUOpsType[mipsOp.f.func] : cop1OpsType[mipsOp.f.rs]) : \
      (mainOpsType[mipsOp.f.op] == COP2) ? ((cop2OpsType[mipsOp.f.rs] == VECTOR) ? cop2VectorOpsType[mipsOp.f.func] : cop2OpsType[mipsOp.f.rs]) : \
      mainOpsType[mipsOp.f.op]

#define getopcode(mipsOp) \
    (mainOpsType[mipsOp.f.op] == SPECIAL) ? specialOps[mipsOp.f.func] : \
    (mainOpsType[mipsOp.f.op] == REGIMM) ? regimmOps[mipsOp.f.rt] : \
    (mainOpsType[mipsOp.f.op] == COP0) ? cop0Ops[mipsOp.f.rs] : \
    (mainOpsType[mipsOp.f.op] == COP1) ? ((cop1OpsType[mipsOp.f.rs] == FPU) ? cop1FPUOps[mipsOp.f.func] : cop1Ops[mipsOp.f.rs]) : \
    (mainOpsType[mipsOp.f.op] == COP2) ? ((cop2OpsType[mipsOp.f.rs] == VECTOR) ? cop2VectorOps[mipsOp.f.func] : cop2Ops[mipsOp.f.rs]) : \
    mainOps[mipsOp.f.op]

#endif /* __NOP_H__ */


